"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""
import logging
import threading
import time
from io import StringIO
from unittest import TestResult, TestCase

from . import save_images
from .log.handler_factory import HandlerFactory
from .retry_on_exception import no_retry_lists, retry_lists


class Result(TestResult):
    """
    定义继承自 unittest.TestResult 的 类。
    这里重写了 unittest.TestResult 的多个方法，比如 startTest(self, test) 等等
    """

    def __init__(self, lang, tries, delay, back_off, max_delay, retry, thread_start_wait):
        super().__init__()
        self.success_count = 0
        self.failure_count = 0
        self.skip_count = 0
        self.error_count = 0
        self.success_set = set()
        self.failure_set = set()
        self.skip_set = set()
        self.error_set = set()
        self.stderr_steams = StringIO()
        self.stderr_steams.write("\n")
        self.stdout_steams = StringIO()
        self.stdout_steams.write("\n")
        self.LANG = lang
        self.tries, self.delay, self.back_off, self.max_delay, self.retry, = tries, delay, back_off, max_delay, retry
        self.thread_start_wait = thread_start_wait
        """
        返回结果是一个5个属性的字典的列表
        (
          result_code (0: success; 1: fail; 2: error; 3: skip),
          testCase_object,
          test_output dict(list(string)),
          image_paths dict,
          tries int,
          style dict
        )
        """
        self.result = []
        self.result_tmp = {}
        self.time = {}

    def startTest(self, test):
        if self.thread_start_wait:
            logging.info(f"测试延迟启动：{self.thread_start_wait}s")
            time.sleep(self.thread_start_wait)
        logging.info((self.LANG == "cn" and "开始测试： {}" or "Start Test: {}").format(test))
        current_id = str(threading.current_thread().ident)
        if current_id in self.result_tmp:
            self.result_tmp[current_id]["tries"] += 1
        else:
            self.result_tmp[current_id] = dict(
                result_code=0,
                testCase_object=test,
                test_output={},
                image_paths={},
                tries=0,
                retry=True,
                style={},
                local_delay=self.delay
            )

        self.time[str(threading.current_thread().ident)] = time.time()
        TestResult.startTest(self, test)

    def stopTest(self, test):
        end_time = time.time()
        logging.info((self.LANG == "cn" and "测试结束： {}" or "Stop Test: {}").format(test))
        logging.info((self.LANG == "cn" and "耗时： {}" or "Duration: {}").format(
            end_time - self.time[str(threading.current_thread().ident)]))
        current_id = str(threading.current_thread().ident)
        tries = self.result_tmp[current_id]["tries"]
        if current_id in save_images.imageList:
            tmp = save_images.imageList.pop(current_id)
            self.result_tmp[current_id]["image_paths"][tries] = tmp
        tmp = HandlerFactory.get_stream_value()
        self.result_tmp[current_id]["test_output"][tries] = tmp

        # 停止重试
        test_name = test.__class__.__name__ + "." + test.__getattribute__("_testMethodName")
        if (not self.retry and test_name not in retry_lists) or (
                self.retry and test_name in no_retry_lists
        ) or self.tries <= tries or self.result_tmp[current_id]["result_code"] in (0, 3):
            self.result_tmp[current_id]["retry"] = False

        if self.result_tmp[current_id]["retry"]:
            # 重试
            if self.result_tmp[current_id]["local_delay"] > self.max_delay:
                self.result_tmp[current_id]["local_delay"] = self.max_delay
            logging.info(
                (self.LANG == "cn" and "等待 {} 秒后重试" or "Retrying in {} seconds...").format(
                    self.result_tmp[current_id]["local_delay"]
                )
            )
            time.sleep(self.result_tmp[current_id]["local_delay"])
            self.result_tmp[current_id]["local_delay"] *= self.back_off
            test(self)
        else:
            # 最后清理
            if current_id in self.success_set:
                self.success_set.remove(current_id)
            if current_id in self.failure_set:
                self.failure_set.remove(current_id)
            if current_id in self.skip_set:
                self.skip_set.remove(current_id)
            if current_id in self.error_set:
                self.error_set.remove(current_id)
            if "retry" in self.result_tmp[current_id]:
                del self.result_tmp[current_id]["retry"]
            if "local_delay" in self.result_tmp[current_id]:
                del self.result_tmp[current_id]["local_delay"]
            # 产生结果
            self.result.append(self.result_tmp.pop(current_id))

    def addSkip(self, test, reason):
        TestResult.addSkip(self, test, reason)

        self._steams_write_doc("Skip", test)

        logging.info((self.LANG == "cn" and "跳过测试： {}\n{}" or "Skip Test: {}\n{}").format(test, reason))

        current_id = str(threading.current_thread().ident)
        self.result_tmp[current_id]["result_code"] = 3
        self.result_tmp[current_id]["style"][self.result_tmp[current_id]["tries"]] = 3
        if current_id not in self.skip_set:
            self.skip_count += 1
            self.skip_set.add(current_id)

    def addSuccess(self, test):
        TestResult.addSuccess(self, test)

        self._steams_write_doc("Pass", test)

        logging.info((self.LANG == "cn" and "测试执行通过： {}" or "Pass Test: {}").format(test))

        current_id = str(threading.current_thread().ident)
        self.result_tmp[current_id]["result_code"] = 0
        self.result_tmp[current_id]["style"][self.result_tmp[current_id]["tries"]] = 0
        if current_id not in self.success_set:
            self.success_count += 1
            self.success_set.add(current_id)
        if current_id in self.failure_set:
            self.failure_count -= 1
        if current_id in self.error_set:
            self.error_count -= 1

    def addError(self, test, err):
        TestResult.addError(self, test, err)
        _, _exc_str = self.errors[-1]

        self._steams_write_doc("Error", test)

        logging.error((self.LANG == "cn" and "测试产生错误： {}\n{}" or "Error Test: {}\n{}").format(test, _exc_str))

        current_id = str(threading.current_thread().ident)
        self.result_tmp[current_id]["result_code"] = 2
        self.result_tmp[current_id]["style"][self.result_tmp[current_id]["tries"]] = 2
        if current_id not in self.error_set:
            self.error_count += 1
            self.error_set.add(current_id)

    def addFailure(self, test, err):
        TestResult.addFailure(self, test, err)
        _, _exc_str = self.failures[-1]

        self._steams_write_doc("Fail", test)

        logging.warning((self.LANG == "cn" and "测试未通过： {}\n{}" or "Failure: {}\n{}").format(test, _exc_str))

        current_id = str(threading.current_thread().ident)
        self.result_tmp[current_id]["result_code"] = 1
        self.result_tmp[current_id]["style"][self.result_tmp[current_id]["tries"]] = 1
        if current_id not in self.failure_set:
            self.failure_count += 1
            self.failure_set.add(current_id)

    def _steams_write_doc(self, result: str, test: TestCase):
        if result == "Pass":
            steams = self.stdout_steams
        else:
            steams = self.stderr_steams

        steams.write(f"{result}\t")
        steams.write(str(test))
        doc = "_testMethodDoc" in test.__dir__() and test.__getattribute__("_testMethodDoc") or ""
        if doc:
            steams.write("\t")
            steams.write(doc.strip().split("\n")[0])
        steams.write("\n")



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""
import logging
import logging.handlers
import sys
import threading
from io import StringIO
from logging import Handler, getLevelName

_LOGGER_FORMAT = "%(asctime)s %(thread)7d %(levelname)8s %(filename)s(%(lineno)d) - %(message)s"


class _StreamHandler(Handler):
    """
    分线程日志流记录
    """

    terminator = '\n'

    def __init__(self, streams=None):
        """
        初始化处理程序
        """
        Handler.__init__(self)
        if streams is None:
            streams = {}
        self.streams = streams

    def flush(self):
        """
        刷新流
        """
        self.acquire()
        steam_id = str(threading.current_thread().ident)
        try:
            if self.streams.get(steam_id) and hasattr(self.streams[steam_id], "flush"):
                self.streams[steam_id].flush()
        finally:
            self.release()

    def emit(self, record):
        """
        记录
        """
        try:
            msg = self.format(record)
            steam_id = str(threading.current_thread().ident)
            if steam_id not in self.streams:
                self.streams[steam_id] = StringIO()
            stream = self.streams[steam_id]
            # issue 35046: merged two stream.writes into one.
            stream.write(msg + self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)

    def __repr__(self):
        level = getLevelName(self.level)
        steam_id = str(threading.current_thread().ident)
        name = getattr(self.streams[steam_id], "name", "")
        if name:
            name += " "
        return f"<{self.__class__.__name__} {name}({level})>"


class InfoOrLessCritical(logging.Filter):
    def filter(self, record):
        return record.levelno < logging.WARNING


class HandlerFactory(object):
    handlers = {}
    streams = {}

    @classmethod
    def get_std_out_handler(cls):
        if "std_out_handler" not in cls.handlers:
            std_out_handler = logging.StreamHandler(sys.stdout)
            std_out_handler.setFormatter(logging.Formatter(_LOGGER_FORMAT))
            std_out_handler.addFilter(InfoOrLessCritical())
            cls.handlers["std_out_handler"] = std_out_handler

        return cls.handlers["std_out_handler"]

    @classmethod
    def get_std_err_handler(cls):
        if "std_err_handler" not in cls.handlers:
            std_err_handler = logging.StreamHandler(sys.stderr)
            std_err_handler.setFormatter(logging.Formatter(_LOGGER_FORMAT))
            std_err_handler.setLevel(logging.WARNING)
            cls.handlers["std_err_handler"] = std_err_handler

        return cls.handlers["std_err_handler"]

    @classmethod
    def get_rotating_file_handler(cls, log_path, max_bytes=0, backup_count=0):
        if "rotating_file_handler" not in cls.handlers:
            cls.handlers["rotating_file_handler"] = {}

        if log_path not in cls.handlers["rotating_file_handler"]:
            rotating_file_handler = logging.handlers.RotatingFileHandler(
                log_path, "a", max_bytes, backup_count, encoding="utf8")
            rotating_file_handler.setLevel(logging.NOTSET)
            rotating_file_handler.setFormatter(logging.Formatter(_LOGGER_FORMAT))
            cls.handlers["rotating_file_handler"][log_path] = rotating_file_handler

        return cls.handlers["rotating_file_handler"][log_path]

    @classmethod
    def get_stream_handler(cls):
        if "rotating_stream_handler" not in cls.handlers:
            rotating_stream_handler = _StreamHandler(cls.streams)
            rotating_stream_handler.setFormatter(logging.Formatter(_LOGGER_FORMAT))
            cls.handlers["rotating_stream_handler"] = rotating_stream_handler

        return cls.handlers["rotating_stream_handler"]

    @classmethod
    def get_stream_value(cls):
        steam_id = str(threading.current_thread().ident)
        if steam_id in cls.streams:
            stream = cls.streams[steam_id].getvalue()
            cls.streams[steam_id].truncate(0)
            cls.streams[steam_id].seek(0)
            return stream
        return ""



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.

本页代码改写：https://github.com/datadriventests/ddt

在其基础上进行改写以符合个性化需求。
"""
import logging
import re
from functools import wraps

from . import retry_on_exception

_DATA_ATTR: str = "%values"  # 存储测试必须运行的数据
_UNPACK_ATTR: str = "%unpack"  # 解包
_index_len: int = 1  # 默认最大的case索引长度

trivial_types = (type(None), bool, int, float, str)


def _is_trivial(value):
    if isinstance(value, trivial_types):
        return True
    elif isinstance(value, (list, tuple)):
        return all(map(_is_trivial, value))
    return False


def unpack(func):
    """
    方法装饰器添加解包特性
    """
    setattr(func, _UNPACK_ATTR, True)
    return func


def data(*values):
    """
    方法装饰器添加到 unittest.TestCase 实例的方法中

    接收位置参数，构成数据驱动
    """
    global _index_len
    _index_len = len(str(len(values)))
    return iter_data(values)


def iter_data(iterable):
    """
    方法装饰器添加到 unittest.TestCase 实例的方法中

    接收可迭代对象，构成数据驱动
    """

    def wrapper(func):
        setattr(func, _DATA_ATTR, iterable)
        return func

    return wrapper


def _mk_test_name(name, value, index: int = 0):
    """
    为测试用例生成一个新名称。

    它将采用原始测试名称，并附加一个序号索引和该值的字符串表示形式，并将结果转换为有效的Python标识符，方法是用`_``替换多余的字符。
    """

    # 在索引之前添加0以保持顺序
    index = f"{index + 1:0{_index_len}}"
    if not _is_trivial(value):
        return f"{name}_{index}"
    try:
        value = str(value)
    except UnicodeEncodeError:
        value = value.encode("ascii", "backslashreplace")
    test_name = f"{name}_{index}_{value}"
    return re.sub(r"\W|^(?=\d)", "_", test_name)


def _feed_data(func, new_name, *args, **kwargs):
    """
    这个内部方法装饰器将测试数据项提供给测试

    """

    @wraps(func)
    def wrapper(self):
        if args:
            logging.info(f"args:\n{args}")
        if kwargs:
            logging.info(f"kwargs:\n{kwargs}")
        return func(self, *args, **kwargs)

    wrapper.__name__ = new_name
    wrapper.__wrapped__ = func

    return wrapper


def _add_test(cls, test_name, func, *args, **kwargs):
    """
    向该类添加一个测试用例

    测试将基于现有的功能，但会给它一个新名称。
    """
    setattr(cls, test_name, _feed_data(func, test_name, *args, **kwargs))
    # 测试重试
    if hasattr(func, retry_on_exception.__retry):
        retry_on_exception.retry_lists.append(cls.__name__ + "." + test_name)
    if hasattr(func, retry_on_exception.__no_retry):
        retry_on_exception.no_retry_lists.append(cls.__name__ + "." + test_name)


def ddt(cls):
    """
    “unittest.TestCase”的子类的类装饰器。

    将这个装饰器应用到测试用例类，然后用 @data 装饰测试方法。

    对于用 @data 装饰的每个方法，这将有效地构建数据驱动。

    测试方法的名称遵循模式 original_test_name_{ordinal}_{data}

    ordinal 是数据参数的位置，从1开始。

    对于数据，使用数据值的字符串表示形式转换为有效的Python标识符
    """
    for name, func in list(cls.__dict__.items()):
        if hasattr(func, _DATA_ATTR):
            for i, v in enumerate(getattr(func, _DATA_ATTR)):
                test_name = _mk_test_name(name, getattr(v, "__name__", v), i)
                if hasattr(func, _UNPACK_ATTR):
                    if isinstance(v, tuple) or isinstance(v, list):
                        _add_test(cls, test_name, func, *v)
                    elif isinstance(v, dict):
                        # 解包字典
                        _add_test(cls, test_name, func, **v)
                else:
                    _add_test(cls, test_name, func, v)
            delattr(cls, name)

            # 测试重试
            func.__qualname__ in retry_on_exception.retry_lists and retry_on_exception.retry_lists.remove(
                func.__qualname__)
            func.__qualname__ in retry_on_exception.no_retry_lists and retry_on_exception.no_retry_lists.remove(
                func.__qualname__)
    return cls



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""
import datetime
import logging
import os
import queue
import random
import time
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Manager
from unittest.suite import TestSuite
from xml.sax import saxutils

from HTMLReport import __author__, __version__
from .tools import save_images
from .tools.log.handler_factory import HandlerFactory
from .tools.result import Result
from .tools.template import TemplateMixin


def _isnotsuite(test):
    """A crude way to tell apart testcases and suites with duck-typing

    :param test:
    :return:
    """
    try:
        iter(test)
    except TypeError:
        return True
    return False


class TestRunner(TemplateMixin, TestSuite):
    """测试执行器"""

    def __init__(self, report_file_name: str = None, log_file_name: str = None, output_path: str = None,
                 title: str = None, description: str = None, tries: int = 0, delay: float = 1, back_off: float = 1,
                 max_delay: float = 120, retry: bool = True, thread_count: int = 1, thread_start_wait: float = 0,
                 sequential_execution: bool = False, lang: str = "cn"):
        """测试执行器

        :param report_file_name: 报告文件名，如果未赋值，将采用“test+时间戳”
        :param log_file_name: 日志文件名，如果未赋值，将采用报告文件名，如果报告文件名也没有，将采用“test+时间戳”
        :param output_path: 报告保存文件夹名，默认“report”
        :param title: 报告标题，默认“测试报告”
        :param description: 报告描述，默认“无测试描述”
        :param tries: 重试次数
        :param delay: 重试延迟间隔，单位为 秒
        :param back_off: 扩展每次重试等待时间的乘数
        :param max_delay: 最大重试等待时间长度，单位为 秒
        :param retry: 如果为 True 表示所有用例遵循重试规则，False 只针对添加了 @retry 用例有效
        :param thread_count: 并发线程数量（无序执行测试），默认数量 1
        :param thread_start_wait: 各线程启动延迟，默认 0 s
        :param sequential_execution: 是否按照套件添加(addTests)顺序执行， 会等待一个addTests执行完成，再执行下一个，默认 False。
                如果用例中存在 tearDownClass ，建议设置为True，否则 tearDownClass 将会在所有用例线程执行完后才会执行。
        :param lang: ("cn", "en") 支持中文与英文报告输出，默认采用中文
        """
        super().__init__()
        self.LANG = lang in ("cn", "en") and lang or "cn"

        self.title = title or (self.LANG == "cn" and self.DEFAULT_TITLE or self.DEFAULT_TITLE_en)
        self.description = description or (
                self.LANG == "cn" and self.DEFAULT_DESCRIPTION or self.DEFAULT_DESCRIPTION_en
        )

        self.thread_count = thread_count
        self.thread_start_wait = thread_start_wait
        self.sequential_execution = sequential_execution

        save_images.report_path = report_path = os.path.join(output_path or "report")
        dir_to = os.path.join(os.getcwd(), report_path)
        if not os.path.exists(dir_to):
            os.makedirs(dir_to)

        random_name = f"test_{time.strftime('%Y_%m_%d_%H_%M_%S')}_{random.randint(1, 999)}"
        report_name = f'{report_file_name or random_name}.html'

        self.log_name = f"{log_file_name or report_file_name or random_name}.log"
        self.path_file = os.path.join(dir_to, report_name)
        self.log_file_name = os.path.join(dir_to, self.log_name)

        logging.getLogger().addHandler(HandlerFactory.get_rotating_file_handler(self.log_file_name))

        self.stopTime = self.startTime = datetime.datetime.now()

        if tries < 0:
            raise ValueError((lang == "cn"
                              and "tries = {} 必须大于或等于 0"
                              or "tries = {} must be greater than or equal to 0").format(tries))
        if delay < 0:
            raise ValueError((lang == "cn"
                              and "delay = {} 必须大于或等于 0"
                              or "delay = {} must be greater than or equal to 0").format(delay))
        if back_off < 1:
            raise ValueError((lang == "cn"
                              and "back_off = {} 必须大于或等于 1"
                              or "back_off = {} must be greater than or equal to 1").format(back_off))
        if max_delay < delay:
            raise ValueError((lang == "cn"
                              and "max_delay = {} 必须大于或等于 delay = {}"
                              or "max_delay = {} must be greater than or equal to delay = {}").format(max_delay, delay))
        self.tries, self.delay, self.back_off, self.max_delay, self.retry = tries, delay, back_off, max_delay, retry
        self.tc_dict = Manager().dict()

    def _thread_pool_executor_test_case(self, tmp_list, result):
        """多线程运行"""
        t_list = []
        with ThreadPoolExecutor(self.thread_count) as pool:
            for test_case in tmp_list:
                if _isnotsuite(test_case):
                    self._tearDownPreviousClass(test_case, result)
                    self._handleModuleFixture(test_case, result)
                    self._handleClassSetUp(test_case, result)
                    result._previousTestClass = test_case.__class__

                    if (getattr(test_case.__class__, "_classSetupFailed", False) or
                            getattr(result, "_moduleSetUpFailed", False)):
                        continue
                while len(t_list) == self.thread_count:
                    [t_list.remove(t) for t in t_list if t.done()]
                t_list.append(pool.submit(test_case, result))

                self.tc_dict[test_case.__class__] -= 1
                if self.tc_dict[test_case.__class__] == 0:
                    self._tearDownPreviousClass(None, result)

        self._handleModuleTearDown(result)

    def run(self, test: TestSuite, debug: bool = False):
        """
        运行给定的测试用例或测试套件。

        :param test: 测试用例或套件
        :param debug: 调试模式
        """
        if debug:
            logging.getLogger().setLevel(logging.DEBUG)

        result = Result(self.LANG, self.tries, self.delay, self.back_off, self.max_delay, self.retry,
                        self.thread_start_wait)
        if self.LANG == "cn":
            logging.info(f"预计并发线程数：{str(self.thread_count)}")
        else:
            logging.info(f"The number of concurrent threads is expected to be {str(self.thread_count)}")

        for test_case in test:
            tmp_class_name = test_case.__class__
            if tmp_class_name not in self.tc_dict:
                self.tc_dict[tmp_class_name] = 0
            else:
                self.tc_dict[tmp_class_name] += 1

        if self.sequential_execution:
            # 执行套件添加顺序
            test_case_queue = queue.Queue()
            test_case_arr = []
            tmp_key = None
            for test_case in test:
                tmp_class_name = test_case.__class__
                if tmp_key == tmp_class_name:
                    test_case_arr.append(test_case)
                else:
                    tmp_key = tmp_class_name
                    if len(test_case_arr) != 0:
                        test_case_queue.put(test_case_arr.copy())
                        test_case_arr.clear()
                    test_case_arr.append(test_case)
            if len(test_case_arr) != 0:
                test_case_queue.put(test_case_arr.copy())
            while not test_case_queue.empty():
                tmp_list = test_case_queue.get()
                self._thread_pool_executor_test_case(tmp_list, result)
        else:
            # 无序执行
            self._thread_pool_executor_test_case(test, result)

        self.stopTime = datetime.datetime.now()
        if result.stdout_steams.getvalue().strip():
            logging.info(result.stdout_steams.getvalue())
        if result.stderr_steams.getvalue().strip():
            logging.error(result.stderr_steams.getvalue())

        if self.LANG == "cn":
            s = "\n测试结束！\n运行时间: {time}\n共计执行用例数量：{count}\n执行成功用例数量：{Pass}" \
                "\n执行失败用例数量：{fail}\n跳过执行用例数量：{skip}\n产生异常用例数量：{error}\n"
        else:
            s = "\nEOT！\nRan {count} tests in {time}\n\nPASS：{Pass}" \
                "\nFailures：{fail}\nSkipped：{skip}\nErrors：{error}\n"
        count = result.success_count + result.failure_count + result.error_count + result.skip_count
        s = s.format(
            time=self.stopTime - self.startTime,
            count=count,
            Pass=result.success_count,
            fail=result.failure_count,
            skip=result.skip_count,
            error=result.error_count
        )
        self._generate_report(result)
        logging.info(s)

        return result

    @staticmethod
    def _sort_result(result_list):
        """unittest不以任何特定的顺序运行。在这里把它们按类分组。"""
        remap = {}
        classes = []
        for dic in result_list:
            n = dic.get("result_code")
            t = dic.get("testCase_object")
            o = dic.get("test_output")
            i = dic.get("image_paths")
            r = dic.get("tries")
            s = dic.get("style")

            cls = t.__class__
            if cls not in remap:
                remap[cls] = []
                classes.append(cls)
            remap[cls].append((n, t, o, i, r, s))

        result = []
        for cls in classes:
            re = remap[cls]
            re = sorted(re, key=lambda x: str(x[1]))
            result.append((cls, re))
        return result

    def _generate_report(self, result):
        """生成报告"""
        generator = f"HTMLReport {__author__} {__version__}"
        stylesheet = self._generate_stylesheet()
        heading = self._generate_heading(result)
        log_file = self._generate_log(self.log_name)
        report = self._create_report(result)

        ending = self._generate_ending()
        js = self._generate_js(result)
        output = self.HTML_TMPL.format(
            title=saxutils.escape(self.title),
            generator=generator,
            stylesheet=stylesheet,
            js=js,
            lang=self.LANG,
            heading=heading,
            log=log_file,
            report=report,
            ending=ending
        )

        with open(self.path_file, "w", encoding="utf8") as report_file:
            report_file.write(output)

    def _generate_stylesheet(self):
        return self.STYLESHEET_TMPL

    def _generate_heading(self, result):
        heading = self.HEADING_TMPL.format(
            title=saxutils.escape(self.title),
            startTime=str(self.startTime)[:19],
            endTime=str(self.stopTime)[:19],
            duration=self.stopTime - self.startTime,
            total=result.success_count + result.failure_count + result.error_count + result.skip_count,
            Pass=result.success_count,
            fail=result.failure_count,
            error=result.error_count,
            skip=result.skip_count,
            description=saxutils.escape(self.description).replace("\n", "<br />"),
        )
        return heading

    def _create_report(self, result):
        rows = []
        nrs = 0
        sorted_result = self._sort_result(result.result)
        for cid, (cls, cls_results) in enumerate(sorted_result):
            np = nf = ne = ns = nr = 0
            for n, t, o, i, r, s in cls_results:
                if n == 0:
                    np += 1
                elif n == 1:
                    nf += 1
                elif n == 2:
                    ne += 1
                elif n == 3:
                    ns += 1
                nr += r
            nrs += nr

            # 格式化类描述
            if cls.__module__ == "__main__":
                name = cls.__name__
            else:
                name = f"{cls.__module__}.{cls.__name__}"
            doc = cls.__doc__ and cls.__doc__.strip().split("\n")[0] or ""
            desc = doc and f"{name}: {doc}" or name
            count = np + nf + ne + ns
            row = self.REPORT_CLASS_TMPL.format(
                style=ne > 0 and "errorClass" or nf > 0 and "failClass" or np > 0 and "passClass" or "skipClass",
                desc=desc,
                count=count,
                Pass=np,
                fail=nf,
                error=ne,
                skip=ns,
                tries=nr,
                statistics=np / (count == 0 and 1 or count),
                cid=f"c{cid + 1}",
            )
            rows.append(row)

            for tid, (n, t, o, i, r, s) in enumerate(cls_results):
                self._generate_report_test(rows, cid, tid, n, t, o, i, r, s)

        count = result.success_count + result.failure_count + result.error_count + result.skip_count
        report = self.REPORT_TMPL.format(
            test_list="".join(rows),
            count=count,
            Pass=result.success_count,
            fail=result.failure_count,
            skip=result.skip_count,
            error=result.error_count,
            tries=nrs,
            statistics=result.success_count / (count == 0 and 1 or count)
        )
        return report

    def _generate_report_test(self, rows, cid, tid, n, t, o, i, r, s):
        # 0: success; 1: fail; 2: error; 3: skip
        tid = (n == 0 and "p" or n == 3 and "s" or n == 2 and "e" or "f") + f"t{cid + 1}.{tid + 1}"
        name = t.id().split(".")[-1]
        doc = "_testMethodDoc" in t.__dir__() and t.__getattribute__("_testMethodDoc") or ""
        doc = doc.strip().split("\n")[0]
        desc = doc and f"{name}: {doc}" or name

        row = self.REPORT_TEST_WITH_OUTPUT_TMPL.format(
            tid=tid,
            tid1=r == 0 and tid + ".1" or tid,
            style=(n == 0 and "passCase" or n == 2 and "errorCase" or
                   n == 1 and "failCase" or n == 3 and "skipCase" or "none"),
            desc=desc,
            status_cn=self.STATUS_cn[n],
            status_en=self.STATUS_en[n],
            count=r + 1,
            tries=r
        )
        rows.append(row)

        for x in range(0, r + 1):
            img_list = ""
            if i.get(x) is not None:
                for img in i.get(x):
                    img_list += self._generate_img(img)

            script = self.REPORT_TEST_OUTPUT_TMPL.format(
                id=f"{tid}.{x + 1}",
                output=saxutils.escape(o.get(x))
            )
            if r == 0:
                row = self.REPORT_TEST_WITH_OUTPUT_SUB_TMPL.format(
                    tid=tid,
                    script=script,
                    img=img_list,
                    r=x + 1,
                )
                rows.append(row)
            else:
                row = self.REPORT_TEST_WITH_OUTPUT_SUB_RETRY_TMPL.format(
                    tid=tid,
                    r=x + 1,
                    style=(s.get(x) == 0 and "passCase" or s.get(x) == 2 and "errorCase" or
                           s.get(x) == 1 and "failCase" or s.get(x) == 3 and "skipCase" or "none"),
                    n=x + 1,
                    status_cn=self.STATUS_cn[s.get(x)],
                    status_en=self.STATUS_en[s.get(x)],
                    count=r + 1
                )
                rows.append(row)

                row = self.REPORT_TEST_WITH_OUTPUT_SUB_TMPL.format(
                    tid=tid,
                    script=script,
                    img=img_list,
                    r=x + 1
                )
                rows.append(row)

    def _generate_ending(self):
        return self.ENDING_TMPL

    def _generate_js(self, result):
        _data = f"""
var chartData_cn = [
  [{result.success_count}, "#1c965b", "通过"],
  [{result.failure_count}, "#ff5722", "失败"],
  [{result.error_count}, "#ff9800", "错误"],
  [{result.skip_count}, "#64b5f6", "跳过"]
];
var chartData_en = [
  [{result.success_count}, "#1c965b", "Passed"],
  [{result.failure_count}, "#ff5722", "Failed"],
  [{result.error_count}, "#ff9800", "Error"],
  [{result.skip_count}, "#64b5f6", "Skipped"]
];
"""
        _js = "function draw() {" + f"""
goChart({"chartData_cn" if self.LANG == "cn" else "chartData_en"});
""" + "}"
        return _data + self.JS + _js

    def _generate_log(self, log_file):
        return self.REPORT_LOG_FILE_TMPL.format(
            log_file=log_file)

    def _generate_img(self, img):
        return self.REPORT_IMG_TMPL.format(img_src=img[0], alt=img[1], title=img[2])



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""
__retry = "%retry"  # 重试
__no_retry = "%no_retry"  # 不重试

retry_lists = [
    # 存储重试列表
]

no_retry_lists = [
    # 存储不重试列表
]


def retry(func):
    """
    方法装饰器添加重试特性
    """
    setattr(func, __retry, True)
    if func.__qualname__ in no_retry_lists:
        raise ValueError("不允许同时存在 @retry 与 @no_retry")
    retry_lists.append(func.__qualname__)
    return func


def no_retry(func):
    """
    方法装饰器添加不重试特性
    """
    setattr(func, __no_retry, True)
    if func.__qualname__ in retry_lists:
        raise ValueError("不允许同时存在 @retry 与 @no_retry")
    no_retry_lists.append(func.__qualname__)
    return func



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.

测试报告生成器
"""
import logging

from .src.tools.log.handler_factory import HandlerFactory

__author__ = '刘士'
# __version__ = '2.0.6a'
__version__ = '2.2.1'

from .src.test_runner import TestRunner
from .src.tools import data_driven as ddt
from .src.tools.retry_on_exception import retry, no_retry
from .src.tools.save_images import addImage

for handler in logging.getLogger().handlers:
    handler: logging
    if not handler.get_name():
        logging.getLogger().removeHandler(handler)

logging.getLogger().setLevel(logging.INFO)
logging.getLogger().addHandler(HandlerFactory.get_std_out_handler())
logging.getLogger().addHandler(HandlerFactory.get_std_err_handler())
logging.getLogger().addHandler(HandlerFactory.get_stream_handler())

__all__ = [
    "addImage",
    "TestRunner",
    "ddt",
    "retry",
    "no_retry",
    "__author__",
    "__version__"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys

if sys.version_info < (3, 7):
    print(u"The minimum support Python 3.7\n支持最低版本 3.7")
    exit(1)
else:
    from setuptools import find_packages
    from setuptools import setup
    from HTMLReport import __version__, __author__

setup(
    name="HTMLReport",
    version=__version__,
    description="Python3 Unittest HTML报告生成器",
    long_description=open("README.md", 'r', encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author=__author__,
    author_email="liushilive@outlook.com",
    url="http://liushilive.github.io",
    project_urls={
        "The report template": "https://liushilive.github.io/report/report/#en",
        "报告样板": "https://liushilive.github.io/report/report/#cn"
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_dir={"HTMLReport": "HTMLReport"},
    include_package_data=True,
    license="Apache 2.0",
    zip_safe=False,
    keywords="HtmlTestRunner test runner html reports unittest",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Testing :: Unit",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: User Interfaces",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: Chinese (Simplified)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy"
    ]
)



"""
Copyright 2017 刘士

Licensed under the Apache License, Version 2.0 (the "License"): you may
not use this file except in compliance with the License. You may obtain
a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""


class TemplateMixin(object):
    # 定义生成HTML结果文件所需要的模板。

    STATUS_cn = {
        0: "通过",
        1: "失败",
        2: "错误",
        3: "跳过",
    }
    STATUS_en = {
        0: "pass",
        1: "fail",
        2: "error",
        3: "skip",
    }

    DEFAULT_TITLE = "测试报告"
    DEFAULT_TITLE_en = "Test Results"
    DEFAULT_DESCRIPTION = "测试描述"
    DEFAULT_DESCRIPTION_en = "Test Description"

    HTML_TMPL = r"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">

<head>
    <title>{title}</title>
    <meta name="generator" content="{generator}"/>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/>
    <style type="text/css" media="screen">{stylesheet}</style>
    <script language="javascript" type="text/javascript">{js}</script>
</head>

<body onload="load()">
<div id="wrapper" class="lang-{lang}">
    <div id="lang">
        <ul>
            <li>
                <a href="#cn" id="lang-cn" title="简体中文">cn</a>
            </li>
            <li>
                <a href="#en" id="lang-en" title="English">en</a>
            </li>
        </ul>
    </div>
    {heading} {log} {report} {ending}
    <div id="popup">
        <div class="bg">
            <img src="" alt=""/>
        </div>
    </div>
</div>
</body>

</html>
"""

    JS = """
function addClass(e, c) {
  if (!isClass(e, c)) {
    if (e.className) {
      e.className = e.className + " " + c;
    } else {
      e.className = c;
    }
  }
}

function delClass(e, c) {
  if (isClass(e, c)) {
    // r = '/(?:^|\s)' + c + '(?!\S)/g';
    let r = new RegExp('(?:^|\\s)' + c + '(?!\\S)', 'g');
    e.className = e.className.replace(r, '');
  }
}

function isClass(e, c) {
  let r = new RegExp('(?:^|\\s)' + c + '(?!\\S)');
  return e.className.match(r);
}

function showCase(level) {
  let trs = document.getElementsByTagName("tr");
  for (let i = 0; i < trs.length; i++) {
    let tr = trs[i];
    let id = tr.id;
    if (id.substr(0, 2) === "st") {
      if (level === 4 || level === 3) {
        delClass(tr, 'hiddenRow');
      } else {
        addClass(tr, 'hiddenRow');
      }
    }
    if (id.substr(0, 2) === "ft") {
      if (level === 4 || level === 2) {
        delClass(tr, 'hiddenRow');
      } else {
        addClass(tr, 'hiddenRow');
      }
    }
    if (id.substr(0, 2) === "pt") {
      if (level === 4 || level === 1) {
        delClass(tr, 'hiddenRow');
      } else {
        addClass(tr, 'hiddenRow');
      }
    }
    if (id.substr(0, 2) === "et") {
      if (level === 4 || level === 5 || level === 2) {
        delClass(tr, 'hiddenRow');
      } else {
        addClass(tr, 'hiddenRow');
      }
    }
    if (id.substr(0, 4) === "div_") {
      addClass(tr, 'hiddenRow');
    }
  }
}

function showClassDetail(cid, count) {
  let id_list = Array(count);
  let toHide = 1;
  for (let i = 0; i < count; i++) {
    let tid0 = "t" + cid.substr(1) + "." + (i + 1);
    let tid = "f" + tid0;
    let tr = document.getElementById(tid);
    if (!tr) {
      tid = "p" + tid0;
      tr = document.getElementById(tid);
      if (!tr) {
        tid = "e" + tid0;
        tr = document.getElementById(tid);
        if (tr === null) {
          tid = "s" + tid0;
          tr = document.getElementById(tid);
        }
      }
    }
    id_list[i] = tid;
    if (tr.className) {
      toHide = 0;
    }
  }
  for (let i = 0; i < count; i++) {
    let tid = id_list[i];
    if (toHide && tid.indexOf("p") !== -1) {
      addClass(document.getElementById(tid), 'hiddenRow');
    } else {
      delClass(document.getElementById(tid), 'hiddenRow');
    }
  }
  let trs = document.getElementsByTagName("tr");
  for (let i = 0; i < trs.length; i++) {
    let tr = trs[i];
    let id = tr.id;
    if (id.substr(0, 4) === "div_") {
      addClass(tr, 'hiddenRow');
    }
  }
}

function showTestDetail(div_id, count, b) {
  let details_div_s = document.getElementsByName(div_id);
  for (let j = 0; j < details_div_s.length; j++) {
    let details_div = details_div_s[j];
    if (isClass(details_div, 'hiddenRow')) {
      delClass(details_div, 'hiddenRow');
    } else {
      addClass(details_div, "hiddenRow");
    }
  }
  for (let i = 1; i <= count; i++) {
    let details_div_s = document.getElementsByName(div_id + '.' + i);
    for (let j = 0; j < details_div_s.length; j++) {
      let details_div = details_div_s[j];
      if (details_div !== undefined) {
        if (b && isClass(details_div, 'hiddenRow')) {
          delClass(details_div, 'hiddenRow');
        } else {
          addClass(details_div, "hiddenRow");
        }
      }
    }
  }
}

function html_escape(s) {
  s = s.replace(/&/g, "&amp;");
  s = s.replace(/</g, "&lt;");
  s = s.replace(/>/g, "&gt;");
  return s;
}

function goChart(dataArr) {

  // 声明所需变量
  var canvas, ctx;
  // 图表属性
  var cWidth, cHeight, cMargin, cSpace;
  // 饼状图属性
  var radius, ox, oy;//半径 圆心
  var tWidth, tHeight;//图例宽高
  var posX, posY, textX, textY;
  var startAngle, endAngle;
  var totleNb;
  // 运动相关变量
  var ctr, numctr, speed;
  //鼠标移动
  var mousePosition = {};

  //线条和文字
  var lineStartAngle, line, textPadding, textMoveDis;

  // 获得canvas上下文
  canvas = document.getElementById("chart");
  if (canvas && canvas.getContext) {
    ctx = canvas.getContext("2d");
  }
  initChart();

  // 图表初始化
  function initChart() {
    // 图表信息
    cMargin = 20;
    cSpace = 40;

    canvas.width = canvas.parentNode.getAttribute("width") * 2;
    canvas.height = canvas.parentNode.getAttribute("height") * 2;
    canvas.style.height = canvas.height / 2 + "px";
    canvas.style.width = canvas.width / 2 + "px";
    cHeight = canvas.height - cMargin * 2;
    cWidth = canvas.width - cMargin * 2;

    //饼状图信息
    radius = cHeight * 2 / 6;  //半径  高度的2/6
    ox = canvas.width / 2 + cSpace;  //圆心
    oy = canvas.height / 2;
    tWidth = 60; //图例宽和高
    tHeight = 20;
    posX = cMargin;
    posY = cMargin;   //
    textX = posX + tWidth + 15
    textY = posY + 18;
    startAngle = endAngle = 90 * Math.PI / 180; //起始弧度 结束弧度
    rotateAngle = 0; //整体旋转的弧度

    //将传入的数据转化百分比
    totleNb = 0;
    new_data_arr = [];
    for (var i = 0; i < dataArr.length; i++) {
      totleNb += dataArr[i][0];
    }
    for (var i = 0; i < dataArr.length; i++) {
      new_data_arr.push(dataArr[i][0] / totleNb);
    }
    totalYNomber = 10;
    // 运动相关
    ctr = 1;//初始步骤
    numctr = 50;//步骤
    speed = 1.2; //毫秒 timer速度

    //指示线 和 文字
    lineStartAngle = -startAngle;
    line = 40;         //画线的时候超出半径的一段线长
    textPadding = 10;  //文字与线之间的间距
    textMoveDis = 200; //文字运动开始的间距
  }

  drawMarkers();
  //绘制比例图及文字
  function drawMarkers() {
    ctx.textAlign = "left";
    for (var i = 0; i < dataArr.length; i++) {
      //绘制比例图及文字
      ctx.fillStyle = dataArr[i][1];
      ctx.fillRect(posX, posY + 40 * i, tWidth, tHeight);
      ctx.moveTo(parseInt(posX)+0.5, parseInt(posY + 40 * i)+0.5);
      ctx.font = 'normal 24px 微软雅黑';    //斜体 30像素 微软雅黑字体
      ctx.fillStyle = dataArr[i][1]; //"#000000";
      var percent = dataArr[i][2] + "：" + parseInt(100 * new_data_arr[i]) + "%";
      ctx.fillText(percent, parseInt(textX)+0.5, parseInt(textY + 40 * i)+0.5);
    }
  };

  //绘制动画
  pieDraw();
  function pieDraw(mouseMove) {

    for (var n = 0; n < dataArr.length; n++) {
      ctx.fillStyle = ctx.strokeStyle = dataArr[n][1];
      ctx.lineWidth = 1;
      var step = new_data_arr[n] * Math.PI * 2; //旋转弧度
      var lineAngle = lineStartAngle + step / 2;   //计算线的角度
      lineStartAngle += step;//结束弧度

      ctx.beginPath();
      var x0 = ox + radius * Math.cos(lineAngle),//圆弧上线与圆相交点的x坐标
        y0 = oy + radius * Math.sin(lineAngle),//圆弧上线与圆相交点的y坐标
        x1 = ox + (radius + line) * Math.cos(lineAngle),//圆弧上线与圆相交点的x坐标
        y1 = oy + (radius + line) * Math.sin(lineAngle),//圆弧上线与圆相交点的y坐标
        x2 = x1,//转折点的x坐标
        y2 = y1,
        linePadding = ctx.measureText(dataArr[n][2]).width + 10; //获取文本长度来确定折线的长度

      ctx.moveTo(parseInt(x0)+0.5, parseInt(y0)+0.5);
      //对x1/y1进行处理，来实现折线的运动
      yMove = y0 + (y1 - y0) * ctr / numctr;
      ctx.lineTo(parseInt(x1)+0.5, parseInt(yMove)+0.5);
      if (x1 <= x0) {
        x2 -= line;
        ctx.textAlign = "right";
        ctx.lineTo(parseInt(x2 - linePadding)+0.5, parseInt(yMove)+0.5);
        ctx.fillText(dataArr[n][2], x2 - textPadding - textMoveDis * (numctr - ctr) / numctr, y2 - textPadding);
      } else {
        x2 += line;
        ctx.textAlign = "left";
        ctx.lineTo(parseInt(x2 + linePadding)+0.5, parseInt(yMove)+0.5);
        ctx.fillText(dataArr[n][2], x2 + textPadding + textMoveDis * (numctr - ctr) / numctr, y2 - textPadding);
      }

      ctx.stroke();

    }

    //设置旋转
    ctx.save();
    ctx.translate(ox, oy);
    ctx.rotate((Math.PI * 2 / numctr) * ctr / 2);

    //绘制一个圆圈
    ctx.strokeStyle = "rgba(0,0,0," + 0.5 * ctr / numctr + ")"
    ctx.beginPath();
    ctx.arc(0, 0, (radius + 20) * ctr / numctr, 0, Math.PI * 2, false);
    ctx.stroke();

    for (var j = 0; j < dataArr.length; j++) {

      //绘制饼图
      endAngle = endAngle + new_data_arr[j] * ctr / numctr * Math.PI * 2; //结束弧度

      ctx.beginPath();
      ctx.moveTo(0, 0); //移动到到圆心
      ctx.arc(0, 0, radius * ctr / numctr, startAngle, endAngle, false); //绘制圆弧

      ctx.fillStyle = dataArr[j][1];
      if (mouseMove && ctx.isPointInPath(mousePosition.x * 2, mousePosition.y * 2)) {
        ctx.globalAlpha = 0.8;
      }

      ctx.closePath();
      ctx.fill();
      ctx.globalAlpha = 1;

      startAngle = endAngle; //设置起始弧度
      if (j == dataArr.length - 1) {
        startAngle = endAngle = 90 * Math.PI / 180; //起始弧度 结束弧度
      }
    }

    ctx.restore();

    if (ctr < numctr) {
      ctr++;
      setTimeout(function () {
        //ctx.clearRect(-canvas.width,-canvas.width,canvas.width*2, canvas.height*2);
        ctx.clearRect(-canvas.width, -canvas.height, canvas.width * 2, canvas.height * 2);
        drawMarkers();
        pieDraw();
      }, speed *= 1.085);
    }
  }

  //监听鼠标移动
  var mouseTimer = null;
  canvas.addEventListener("mousemove", function (e) {
    e = e || window.event;
    if (e.offsetX || e.offsetX == 0) {
      mousePosition.x = e.offsetX;
      mousePosition.y = e.offsetY;
    } else if (e.layerX || e.layerX == 0) {
      mousePosition.x = e.layerX;
      mousePosition.y = e.layerY;
    }

    clearTimeout(mouseTimer);
    mouseTimer = setTimeout(function () {
      ctx.clearRect(0, 0, canvas.width, canvas.height);
      drawMarkers();
      pieDraw(true);
    }, 10);
  });

}

function load() {
  let el_wrapper = document.getElementById('wrapper');
  document.getElementById('lang-cn').onclick = function () {
    el_wrapper.className = 'lang-cn';
    goChart(chartData_cn);
  };
  document.getElementById('lang-en').onclick = function () {
    el_wrapper.className = 'lang-en';
    goChart(chartData_en);
  };

  let nav_lang = (location.hash || '').replace(/#/, '');
  if (nav_lang === 'cn' || nav_lang === 'en') {
    el_wrapper.className = 'lang-' + nav_lang;
  }

  let images = document.getElementsByClassName("pic");
  let lens = images.length;
  let popup = document.getElementById("popup");

  function show(event) {
    event = event || window.event;
    let target = document.elementFromPoint(event.clientX, event.clientY);
    showBig(target.src, target.title, target.alt);
  }

  for (let i = 0; i < lens; i++) {
    images[i].onclick = show;
  }
  popup.onclick = function () {
    popup.getElementsByTagName("img")[0].src = "";
    popup.getElementsByTagName("img")[0].title = "";
    popup.getElementsByTagName("img")[0].alt = "";
    popup.style.display = "none";
    popup.style.zIndex = "-1";
  };

  function showBig(src, title, alt) {
    popup.getElementsByTagName("img")[0].src = src;
    popup.getElementsByTagName("img")[0].title = title;
    popup.getElementsByTagName("img")[0].alt = alt;
    popup.style.display = "block";
    popup.style.zIndex = "999999";
  }

  draw()
}
"""

    STYLESHEET_TMPL = """
    body {
    font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", "PingFang SC", "Hiragino Sans GB", "Microsoft YaHei", "Helvetica Neue", Helvetica, Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
    font-size: 14px;
}

pre {
    word-wrap: break-word;
    word-break: break-all;
    overflow: auto;
    white-space: pre-wrap
}

h1 {
    font-size: 16pt;
    color: gray
}

.heading {
    margin-top: 0;
    margin-bottom: 1ex
}

.heading .attribute {
    margin-top: 1ex;
    margin-bottom: 0
}

.heading .description {
    margin-top: 4ex;
    margin-bottom: 6ex
}

a.popup_link:hover {
    color: red
}

.popup_window {
    display: block;
    position: relative;
    left: 0;
    top: 0;
    padding: 10px;
    background-color: #E6E6D6;
    text-align: left;
    font-size: 13px
}

.popup_retry_window {
    padding-left: 50px;
}

#show_detail_line {
    margin-top: 3ex;
    margin-bottom: 1ex
}

#result_table {
    width: 100%;
    border-collapse: collapse;
    border: 1px solid #777
}

#header_row {
    color: #fff;
    background-color: #777
}

#result_table td {
    border: 1px solid #777;
    padding: 2px;
}

#result_table td:nth-child(n+2) {
    min-width: 70px;
    width: 100%
}

#result_table td:nth-child(n+3) {
    text-align: center;
}

#result_table td:first-of-type {
    text-align: center;
    min-width: 60px;
}

#total_row {
    font-weight: bold
}

.passClass,
.failClass,
.errorClass,
.skipClass {
    font-weight: bold
}

.passCase {
  background-color: #d0e9c6
}

.failCase {
  background-color: #ebcccc
}

.errorCase {
  background-color: #faf2cc
}

.skipCase {
  background-color: #c4e3f3
}

.hiddenRow {
    display: none
}

.testcase {
    margin-left: 2em
}

#popup {
    position: fixed;
    left: 0;
    top: 0;
    width: 100%;
    height: 100%;
    text-align: center;
    display: none
}

#popup .bg {
    background-color: rgba(0, 0, 0, .5);
    width: 100%;
    height: 100%
}

#popup img {
    max-width: 100%;
    max-height: 100%;
    margin: auto;
    position: absolute;
    top: 0;
    left: 0;
    bottom: 0;
    right: 0;
}

img.pic {
    cursor: pointer;
    width: auto;
    height: auto;
    max-width: 100%;
    max-height: 100%;
}

#wrapper {
    margin: 0 auto;
    border-top: solid 2px #666;
}

#wrapper .lang-en {
    display: none;
}

#wrapper.lang-cn p.lang-cn {
    display: block;
}

#wrapper.lang-cn span.lang-cn {
    display: inline;
}

#wrapper.lang-cn .lang-en {
    display: none;
}

#wrapper.lang-en .lang-cn {
    display: none;
}

#wrapper.lang-en p.lang-en {
    display: block;
}

#wrapper.lang-en span.lang-en {
    display: inline;
}

#lang ul {
    float: right;
    margin: 0;
    padding: 2px 10px 4px 10px;
    background: #666;
    border-radius: 0 0 4px 4px;
    list-style: none;
}

#lang li {
    margin: 0;
    float: left;
    padding: 0 5px;
}

#lang li a {
    display: block;
    width: 24px;
    height: 24px;
    text-indent: -4em;
    overflow: hidden;
    background: transparent url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAIGNIUk0AAHolAACAgwAA+f8AAIDpAAB1MAAA6mAAADqYAAAXb5JfxUYAAAHiSURBVHja1Ja/jtNAEMZ/Y2/icBdxXAMSEu/A1dBR0NJQ8CS8AQ0tb4CQgEegPgQFOh7ixJUX4vgSx96ZoUgOO3+KRDgFX7Or0Wg+f7PzeVfcnUMi4cA4OIEAARgAvY5r10AZgOGvl69Gkm4Xk9w3fJTg9f4MDz9+OA3AsSTC4OmThaQE3Bp9w+eRmy+hie2I8us3gOMABFNFkjlW5PTPIvOLAO7YVMjfC/Sd4YuK4nOGuyMiABv7v6pP7mKmACEAeK1YPuPoWU52FgkPUiaf+ngFDjCD+Q/Fproo1vrSbUPuvR4eF7kBwDRi4ynlzxkyUMrvKTZabbrPFb9Jd2qPh7BK4DGiRYFeTJmdC8nAsVKaUes72eOK6Xm2G0GaYhpXCTzPsXEBgOZN8unrktHbAddvAKrdCESwqmoItI74eILlkw0bjt4Zltdg+5hL8NhSYLGmurrCxuPN7Mv951+LAh1kLQWxBlUw68bDGtEqaStQiB0SRMWlbh1yXWPu+MIc/wzTiC0dslBQR0TArfWPwJdr21KyttLKaeJijvmaD0gTMF/z57pPt8W37E1xaylwU0iE5OhON2fgjreMVmuMXC/ntus7QYAT4BFwr+Piv4HL2xstu21Xh4jAXP77V8WfAQAixA0rudAk0AAAAABJRU5ErkJggg==") no-repeat 50% 50%;
}

#lang li a#lang-en {
    background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABgAAAAYCAYAAADgdz34AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAIGNIUk0AAHolAACAgwAA+f8AAIDpAAB1MAAA6mAAADqYAAAXb5JfxUYAAAIWSURBVHja1JY/SBthGIefu1xqS6K20KFDy0kopUiHmphIByUZotRAIZOTWZzFpYtbB0uh6KJTIdQhi9pBSwmmCOpgoUSKFItTh4AU6tCr16Rn5P58XZocDrlYuAz9wfHAcbzv9/2+932/k4QQdFIyHVbHE0iAAlwFgj7HNoG6AoRzudc/A4F/28yL2l7bb269yd9QgJAsS8zMjFIufyWRuHspXqtbnsHrH8oAIQlQJyfzlaGhCNFohJ2dI1Kp/iZ3d49IJvsvvJckmJ197JlACIEsy30KgGUJBgcjFIufSacfsLnpza2tL/x4+qx15fR0Uz84hL8HjG1blEqHJJP9bGx8IpMZ8CSAMIzWq1cUhO24CSzLYWTkPisrH8lm46yuenN9fZ+br156WmRZFgQLjR3YrK2VyWSiFAp7TEw88iTAyZNca4t6e6h/P3EbzTRtxscfks9vk83G27JaPcOuVls/v6o4pltlajo9L1KpebG8vC9isbm2jMXmRDsZhiEAVWn4NTU1ysJCkenpMRYXS55cWnrPcSThUUVhzrquNEeFOjz8vOI4CrXa+aU7+d3p29YJusMYwQD3Drb7AFRd14Xf0nXdtehbfAxdkhG13/5M0HCImiTcPhC2BVIAHMefOWrbCNxYqqZpvlukaVrTIrNye4CK1JH7xpSAXuAOcN3n4KfAceNG62qch4+ygHPpv/+r+DMAXV79BpyNnBoAAAAASUVORK5CYII=");
}

.figure_ul {
    text-align: center;
    padding: 0;
}

.figure_li {
    width: 30em;
    list-style: none;
    display: inline-block;
    vertical-align: baseline;
}

tr {
    height: 2em;
}
"""

    HEADING_TMPL = r"""<div class='heading'>
<h1>{title}</h1>
<table>
<tr><td style="width: 100%; vertical-align: top;">
  <p class='attribute'>
    <strong>
      <span class="lang-cn">启动时间：</span>
      <span class="lang-en">Start Time:</span>
    </strong> {startTime}
  </p>
  <p class='attribute'>
    <strong>
      <span class="lang-cn">结束时间：</span>
      <span class="lang-en">End Time:</span>
    </strong> {endTime}
  </p>
  <p class='attribute'>
    <strong>
      <span class="lang-cn">运行时长：</span>
      <span class="lang-en">Duration:</span>
    </strong> {duration}
  </p>
  <p class='attribute'>
    <strong>
      <span class="lang-cn">结果：</span>
      <span class="lang-en">Status:</span>
    </strong>
    <span class="lang-cn">合计：</span>
    <span class="lang-en">Total:</span>{total}&nbsp;&nbsp;&nbsp;&nbsp;
    <span class="lang-cn">通过：</span>
    <span class="lang-en">Passed:</span>{Pass}&nbsp;&nbsp;&nbsp;&nbsp;
    <span class="lang-cn">失败：</span>
    <span class="lang-en">Failed:</span>{fail}&nbsp;&nbsp;&nbsp;&nbsp;
    <span class="lang-cn">错误：</span>
    <span class="lang-en">Error:</span>{error}&nbsp;&nbsp;&nbsp;&nbsp;
    <span class="lang-cn">跳过：</span>
    <span class="lang-en">Skipped:</span>{skip}&nbsp;&nbsp;&nbsp;&nbsp;
  </p>
  <p class='description'>{description}</p>
  </td>
  <td>
    <div height="400" width="600">
      <canvas id="chart" style="border: 1px solid #A4E2F9;"> 你的浏览器不支持HTML5 canvas </canvas>
    </div>
  </td>
</tr>
</table>
</div>"""

    REPORT_TMPL = r"""<p id='show_detail_line'>筛选
    <a href='javascript:showCase(0)'>
        <span class="lang-cn">摘要</span>
        <span class="lang-en">Summary</span>
    </a>
    <a href='javascript:showCase(1)'>
        <span class="lang-cn">通过</span>
        <span class="lang-en">Pass</span>
    </a>
    <a href='javascript:showCase(2)'>
        <span class="lang-cn">失败</span>
        <span class="lang-en">Fail</span>
    </a>
    <a href='javascript:showCase(5)'>
        <span class="lang-cn">异常</span>
        <span class="lang-en">Error</span>
    </a>
    <a href='javascript:showCase(3)'>
        <span class="lang-cn">跳过</span>
        <span class="lang-en">Skip</span>
    </a>
    <a href='javascript:showCase(4)'>
        <span class="lang-cn">全部</span>
        <span class="lang-en">All</span>
    </a>
</p>
<table id='result_table'>
    <tr id='header_row'>
        <th>
            <span class="lang-cn">序号</span>
            <span class="lang-en">NO</span>
        </th>
        <th>
            <span class="lang-cn">测试组/测试用例</span>
            <span class="lang-en">Test Group/Test case</span>
        </th>
        <th>
            <span class="lang-cn">计数</span>
            <span class="lang-en">Count</span>
        </th>
        <th>
            <span class="lang-cn">通过</span>
            <span class="lang-en">Passed</span>
        </th>
        <th>
            <span class="lang-cn">失败</span>
            <span class="lang-en">Failed</span>
        </th>
        <th>
            <span class="lang-cn">错误</span>
            <span class="lang-en">Erroneous</span>
        </th>
        <th>
            <span class="lang-cn">跳过</span>
            <span class="lang-en">Skipped</span>
        </th>
        <th>
            <span class="lang-cn">统计</span>
            <span class="lang-en">Statistics</span>
        </th>
        <th>
            <span class="lang-cn">重试</span>
            <span class="lang-en">Tries</span>
        </th>
        <th>
            <span class="lang-cn">查看</span>
            <span class="lang-en">View</span>
        </th>
    </tr>
    {test_list}
    <tr id='total_row'>
        <td>&nbsp;</td>
        <td>
            <span class="lang-cn">合计</span>
            <span class="lang-en">Total</span>
        </td>
        <td>{count}</td>
        <td class="passCase">{Pass}</td>
        <td class="failCase">{fail}</td>
        <td class="errorCase">{error}</td>
        <td class="skipCase">{skip}</td>
        <td style="text-align:right;">{statistics:.2%}</td>
        <td>{tries}</td>
        <td>&nbsp;</td>
    </tr>
</table>
"""

    REPORT_CLASS_TMPL = r"""<tr class='{style}'>
    <td>{cid}</td>
    <td>{desc}</td>
    <td>{count}</td>
    <td class="passCase">{Pass}</td>
    <td class="failCase">{fail}</td>
    <td class="errorCase">{error}</td>
    <td class="skipCase">{skip}</td>
    <td style="text-align:right;">{statistics:.2%}</td>
    <td>{tries}</td>
    <td>
        <a href="javascript:showClassDetail('{cid}',{count})">
            <span class="lang-cn">细节</span>
            <span class="lang-en">Detail</span>
        </a>
    </td>
</tr>
"""

    REPORT_TEST_WITH_OUTPUT_TMPL = r"""<tr id='{tid}'>
<td>{tid}</td>
    <td class='{style}' colspan='7'>
        <div class='testcase'>{desc}</div>
    </td>
    <td class='{style}'>
        <div class='testcase' style="margin-left: auto;">{tries}</div>
    </td>
    <td class='{style}' align='center'>
        <a class="popup_link" onfocus='this.blur();' href="javascript:showTestDetail('div_{tid1}',{count}, false)">
            <span class="lang-cn">{status_cn}</span>
            <span class="lang-en">{status_en}</span>
        </a>
    </td>
</tr>
"""

    REPORT_TEST_WITH_OUTPUT_SUB_RETRY_TMPL = """<tr id='div_{tid}.{r}' name='div_{tid}' class="hiddenRow">
    <td>{tid}.{r}</td>
    <td colspan='7' class='{style}'>
        <div class='popup_retry_window'>
            <span class="lang-cn"> 第 {n} 次尝试 </span>
            <span class="lang-en"> Try {n} </span>
        </div>
    </td>
    <td class='{style}' align='center'>
        <a class="popup_link" onfocus='this.blur();'
           href="javascript:showTestDetail('div_{tid}.{r}',0, true)">
            <span class="lang-cn">{status_cn}</span>
            <span class="lang-en">{status_en}</span>
        </a>
    </td>
</tr>"""

    REPORT_TEST_WITH_OUTPUT_SUB_TMPL = """<tr id='div_S_{tid}.{r}' name='div_{tid}.{r}' class="hiddenRow">
    <td colspan='10'>
        <div class="popup_window">
            <div style='text-align: right; color:red;cursor:pointer'
                 onclick="document.getElementById('div_S_{tid}.{r}').className = 'hiddenRow'">
                <a onfocus='this.blur();'>[x]</a>
            </div>
            <pre>{script}</pre>
            <div><ul class='figure_ul'>{img}<ul></div>
        </div>
    </td>
</tr>"""

    REPORT_TEST_OUTPUT_TMPL = r"""{id}:
{output}
"""

    ENDING_TMPL = r"""<div id='ending'>&nbsp;</div>"""

    REPORT_LOG_FILE_TMPL = r"""<a href='{log_file}'>
    <span class="lang-cn">下载日志文件</span>
    <span class="lang-en">Download log file</span>
</a>
"""

    REPORT_IMG_TMPL = r"""<li class="figure_li">
    <figure>
        <img class="pic" src='{img_src}' title='{alt}' alt='{title}'/>
    </figure>
    {title}
</li>
"""
